import csv
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path

import xlrd  # type: ignore[import-untyped]
from dateutil import parser
from rich.console import Console

from .movements import Movement, Movements

console = Console()


class ImportSource(str, Enum):
    ing = "ing"
    revolut = "revolut"
    paypal = "paypal"


class Importer:
    """Class to import movements from different sources"""

    @staticmethod
    def import_files(source: ImportSource, files: list[Path]):
        movements = Movements.load()
        for path in files:
            if not path.is_file():
                console.print(f"[ERROR] The file {path} does not exists")
                continue
            match source:
                case ImportSource.ing:
                    new_movements = Importer.parse_ing_file(path)
                case ImportSource.revolut:
                    new_movements = Importer.parse_revolut_file(path)
                case ImportSource.paypal:
                    Importer.enrich_with_paypal(movements, path)
                    # No need to continue, the paypal method does everything
                    continue
                case _:
                    console.print(f"[ERROR] The source {source} is not supported")
                    return
            # Consume the movements from the generator
            added = 0
            for move in new_movements:
                if move not in movements.movements:
                    movements.movements.append(move)
                    added += 1
            console.print(f"Imported {added} movements from file {path} using {source.value} importer")
        movements.save()

    @staticmethod
    def parse_ing_file(file_name: Path):
        """Parse movements from an XLS file exported from ING"""
        book = xlrd.open_workbook(file_name)
        sh = book.sheet_by_index(0)
        for rx in range(sh.nrows):
            # Seek the movements
            cell = sh[rx]
            if cell[0].ctype != xlrd.XL_CELL_DATE:
                continue

            # Extract a date from the cell
            movement_date = xlrd.xldate.xldate_as_datetime(cell[0].value, 0).date()
            yield Movement(movement_date, cell[1].value, cell[2].value, cell[3].value, cell[6].value)

    @staticmethod
    def parse_revolut_file(file_name: Path):
        """Parse movements from a CSV file exported from Revolut"""
        reader = csv.DictReader(file_name.open(), delimiter=",")
        for row in reader:
            if row["Type"] in ("CARD_PAYMENT", "Card Payment"):
                # We are only interested in card payments
                continue
            movement_date = parser.parse(row["Started Date"]).date()
            yield Movement(movement_date, "Unknown", "Unknown", row["Description"], float(row["Amount"]))

    @staticmethod
    def enrich_with_paypal(movements: Movements, file_name: Path):
        """ "Enrich movements with PayPal information"""
        reader = csv.DictReader(file_name.open(encoding="utf-8-sig"), delimiter=",")
        paypal_movements = {}
        for row in reader:
            day = datetime.strptime(row["Date"], "%d/%m/%Y").date()  # noqa: DTZ007
            value = float(row["Amount"])
            if row["Status"] != "Completed" or not row["Name"]:
                # We are only interested in completed movements
                continue
            # The paypal transaction seems to be quite delayed sometimes
            for _ in range(6):
                paypal_movements[(day, value)] = row["Name"]
                day = day + timedelta(days=1)
        enriched_movements = 0
        for mv in movements.movements:
            if not mv.paypal_info and (mv.date, mv.value) in paypal_movements:
                mv.paypal_info = paypal_movements[(mv.date, mv.value)]
                enriched_movements += 1
        console.print(f"Enriched {enriched_movements} movements with PayPal information")



from rich.console import Console
from rich.table import Column, Table

from .bills import RegularBills
from .movements import Movements

console = Console()


class Printer:
    @staticmethod
    def list_movements_per_month(movements: Movements):
        table = Table(Column(header="Date", style="blue"), Column(header="Import (€)", justify="right", style="green"))
        movs_per_month = movements.get_movements_per_month()
        for k, v in movs_per_month.items():
            table.add_row(str(k), str(round(v, 2)))
        console.print(table)

    @staticmethod
    def list_movements(movements: Movements):
        table = Table(
            Column(header="Date", style="blue"),
            Column(header="Category", style="green"),
            Column(header="Subcategory", style="cyan"),
            Column(header="Description"),
            Column(header="Import (€)", justify="right", style="green"),
        )

        for m in movements.movements:
            description = m.description
            if m.paypal_info:
                description = f"Paypal ({m.paypal_info})"
            table.add_row(str(m.date), m.category, m.subcategory, description, str(m.value))
        console.print(table)
        console.print(f"Printed {len(table.rows)} movements")

    @staticmethod
    def list_bills(bills: RegularBills):
        table = Table(
            Column(header="First Date", style="blue"),
            Column(header="Last Date", style="blue"),
            Column(header="Category", style="green"),
            Column(header="Description"),
            Column(header="Hits"),
            Column(header="Last import (€)", justify="right", style="green"),
            Column(header="Total (€)", justify="right", style="green"),
        )

        for e in bills.bills.values():
            if e.ignore:
                continue
            table.add_row(
                str(e.start_date),
                str(e.info.date),
                e.info.category,
                e.info.description,
                str(e.hits),
                str(-round(e.info.value, 2)),
                str(-round(e.total, 2)),
            )
        console.print(table)
        console.print(f"Printed {len(table.rows)} fixed expenses")



import datetime
from dataclasses import dataclass

from rich.console import Console

from .movements import Movement, Movements

console = Console()


@dataclass
class RegularBill:
    info: Movement  # Store last movement for common information
    start_date: datetime.date
    hits: int
    total: float
    ignore: bool = False


@dataclass
class RegularBills:
    """This class implements a very simplistic heuristic to try to find regular bills"""

    bills: dict[str, RegularBill]

    def __init__(self, moves: Movements):
        def valid_expense(e):
            period = (e.info.date - e.start_date) / e.hits
            return 22 < period.days < 32 and not e.ignore  # noqa: PLR2004

        moves.movements = [m for m in moves.movements if m.value < 0]
        moves.sort()
        expenses = {}

        for m in moves.movements:
            if m.description not in expenses:
                expenses[m.description] = RegularBill(m, m.date, 1, m.value)
            else:
                e = expenses[m.description]
                if e.hits == 1 and m.value != e.info.value:
                    e.ignore = True
                e.hits += 1
                e.info = m
                e.total += m.value

        self.bills = {k: v for (k, v) in expenses.items() if valid_expense(v)}



from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from pathlib import Path

import unidecode
import yaml
from fastclasses_json import dataclass_json
from rich.console import Console

config_path = Path("~/.ngtonic").expanduser()
movements_db = Path(f"{config_path}/movements.json")
user_config = Path(f"{config_path}/config.yaml")
console = Console()


@dataclass_json
@dataclass
class Movement:
    date: date
    category: str
    subcategory: str
    description: str
    value: float
    paypal_info: str | None = None


def fuzzy_match(needle, hay):
    def normalize(s):
        return unidecode.unidecode(s.lower())

    return normalize(needle) in normalize(hay)


class MovementFilterType(Enum):
    CATEGORY = "category"
    SUBCATEGORY = "subcategory"
    DESCRIPTION = "description"
    INCOME = "income"
    EXPENSE = "expense"
    VALUE = "value"
    START_DATE = "start_date"
    END_DATE = "end_date"


@dataclass
class MovementFilter:
    filter_type: MovementFilterType
    filter_value: str | datetime | None = None

    def is_match(self, movement: Movement):  # noqa: PLR0911
        match self.filter_type:
            case MovementFilterType.CATEGORY:
                return fuzzy_match(self.filter_value, movement.category)
            case MovementFilterType.SUBCATEGORY:
                return fuzzy_match(self.filter_value, movement.subcategory)
            case MovementFilterType.DESCRIPTION:
                return fuzzy_match(self.filter_value, movement.description)
            case MovementFilterType.INCOME:
                return movement.value > 0
            case MovementFilterType.EXPENSE:
                return movement.value < 0
            case MovementFilterType.VALUE:
                return movement.value == float(self.filter_value)  # type: ignore[arg-type]
            case MovementFilterType.START_DATE:
                return movement.date >= self.filter_value.date()  # type: ignore[union-attr]
            case MovementFilterType.END_DATE:
                return movement.date <= self.filter_value.date()  # type: ignore[union-attr]
            case _:
                console.print(f"[ERROR] Unknown filter type: {self.filter_type}")
                return False


@dataclass_json
@dataclass
class Movements:
    movements: list[Movement]

    @staticmethod
    def load():
        if not movements_db.is_file():
            return Movements([])

        with movements_db.open() as f:
            return Movements.from_json(f.read())

    def save(self):
        self.sort()
        if not config_path.is_dir():
            config_path.mkdir()
        with movements_db.open("w+") as f:
            jm = self.to_json()
            f.write(jm)

    def filter(self, filters: list[MovementFilter]):
        self.movements = [m for m in self.movements if all(f.is_match(m) for f in filters)]

        if user_config.is_file():
            with user_config.open() as f:
                config = yaml.safe_load(f)
                self.filter_exclusions(config["excluded_movements"])

        # Sort by date and calculate balance over time
        self.sort()

    def filter_exclusions(self, exclusions):
        def match_exclusion(exclusion, m):
            exclusion_filters = []
            for field, value in exclusion.items():
                exclusion_filters.append(MovementFilter(MovementFilterType(field), value))
            return all(f.is_match(m) for f in exclusion_filters)

        def should_exclude(m):
            return any(match_exclusion(e, m) for e in exclusions)

        self.movements = [m for m in self.movements if not should_exclude(m)]

    def sort(self):
        self.movements.sort(key=lambda m: m.date)

    def get_movements_per_month(self):
        movs_per_month = {}
        for m in self.movements:
            month = date(m.date.year, m.date.month, 1)
            if month not in movs_per_month:
                movs_per_month[month] = 0
            movs_per_month[month] += m.value
        return movs_per_month

    def get_balance_over_time(self):
        moves = {}
        n = 0
        for m in self.movements:
            n += m.value
            moves[m.date] = n
        return moves



from lazy_import import lazy_module  # type: ignore[import-untyped]

from .movements import Movements

# This module is very slow so we lazy load them instead
plt = lazy_module("matplotlib.pyplot")


def running_mean(values, window):
    # Not the fastest mean but faster than importing numpy :-|
    cumsum = [0]
    moving_aves = values[: window - 1]
    for i, n in enumerate(values, 1):
        cumsum.append(cumsum[i - 1] + n)
        if i >= window:
            moving_ave = (cumsum[i] - cumsum[i - window]) / window
            moving_aves.append(moving_ave)
    return moving_aves


class Plotter:
    @staticmethod
    def show_movements_per_month(movements: Movements):
        moves_per_month = movements.get_movements_per_month()
        x = moves_per_month.keys()
        y = list(moves_per_month.values())

        y_mean = running_mean(y, 5)
        fig, ax = plt.subplots()
        ax.bar(x, y, width=25)
        plt.plot(x, y_mean, "r.-", label="Moving average")
        plt.xlabel("Time")
        plt.ylabel("Balance (€)")
        plt.title("Balance per month")
        plt.gcf().autofmt_xdate()
        plt.show()

    @staticmethod
    def show_balance_over_time(movements: Movements):
        moves_per_day = movements.get_balance_over_time()
        x = moves_per_day.keys()
        y = moves_per_day.values()

        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.fill_between(x, y, alpha=0.5)
        ax.plot(x, y, color="k", linewidth=0.5)
        plt.xlabel("Time")
        plt.ylabel("Balance (€)")
        plt.title("Balance over time")
        plt.show()



#! /usr/bin/env python

from datetime import datetime
from pathlib import Path
from typing import Annotated

import typer
from rich.console import Console

from .bills import RegularBills
from .importer import Importer, ImportSource
from .movements import MovementFilter, MovementFilterType, Movements
from .plotter import Plotter
from .printer import Printer

app = typer.Typer(no_args_is_help=True)
console = Console()


def abort(message: str):
    console.print(f"[ERROR] {message}", style="bold red")
    raise typer.Exit(1)


def load_and_filter_movements(categories, descriptions, incomes, expenses, start_date, end_date):
    m = Movements.load()
    if not m.movements:
        abort("The movements file is empty, you need to import some data first")

    movement_filters = []
    if categories:
        movement_filters += [MovementFilter(MovementFilterType.CATEGORY, value) for value in categories]
    if descriptions:
        movement_filters += [MovementFilter(MovementFilterType.DESCRIPTION, value) for value in descriptions]
    if incomes:
        movement_filters.append(MovementFilter(MovementFilterType.INCOME))
    if expenses:
        movement_filters.append(MovementFilter(MovementFilterType.EXPENSE))
    if start_date:
        movement_filters.append(MovementFilter(MovementFilterType.START_DATE, start_date))
    if end_date:
        movement_filters.append(MovementFilter(MovementFilterType.END_DATE, end_date))
    m.filter(movement_filters)

    if not m.movements:
        abort("You filter out all movements, chech your filters")
    return m


# Definition of some parameters shared in several functions
Category = Annotated[list[str] | None, typer.Option("--category", "-c", help="Filter by category")]
Description = Annotated[list[str] | None, typer.Option("--description", "-d", help="Filter by description")]
Incomes = Annotated[bool, typer.Option("--incomes", "-i", help="Show only incomes (positive transactions)")]
Expenses = Annotated[bool, typer.Option("--expenses", "-e", help="Show only incomes (positive transactions)")]
StartDate = Annotated[datetime | None, typer.Option("--from", "-f", help="Show transactions after this date")]
EndDate = Annotated[datetime | None, typer.Option("--to", "-t", help="Show transactions before this date")]
MonthGroup = Annotated[bool, typer.Option("--group-by-month", "-m", help="Group transactions by month")]
ConfigFile = Annotated[str, typer.Option("--config", "-o", help="Config file path")]


@app.command("import")
def import_movements(source: ImportSource, files: list[Path]):
    """Import movements files to the internal storage"""
    Importer.import_files(source, files)


@app.command("list")
def list_movements(
    categories: Category = None,
    descriptions: Description = None,
    incomes: Incomes = False,
    expenses: Expenses = False,
    start_date: StartDate = None,
    end_date: EndDate = None,
    month_group: MonthGroup = False,
):
    """Show a table listing the movements with optional filtering"""
    movements = load_and_filter_movements(categories, descriptions, incomes, expenses, start_date, end_date)
    if month_group:
        Printer.list_movements_per_month(movements)
    else:
        Printer.list_movements(movements)


@app.command()
def balance_plot(
    categories: Category = None,
    descriptions: Description = None,
    incomes: Incomes = False,
    expenses: Expenses = False,
    start_date: StartDate = None,
    end_date: EndDate = None,
):
    """Plot the evolution of the balance over time"""
    movements = load_and_filter_movements(categories, descriptions, incomes, expenses, start_date, end_date)
    Plotter.show_balance_over_time(movements)


@app.command()
def month_plot(
    categories: Category = None,
    descriptions: Description = None,
    incomes: Incomes = False,
    expenses: Expenses = False,
    start_date: StartDate = None,
    end_date: EndDate = None,
):
    """Plot the movements grouped per month"""
    movements = load_and_filter_movements(categories, descriptions, incomes, expenses, start_date, end_date)
    Plotter.show_movements_per_month(movements)


@app.command()
def find_bills():
    """Use a simple heuristic to find regular bills, like subscriptions"""
    movements = Movements.load()
    fe = RegularBills(movements)
    Printer.list_bills(fe)


if __name__ == "__main__":
    app()
